
/* writeasc.c - dump the internal database to a ASC file
 *            - written by Shawn Knight - 11/30/93
 *            - based on code written by Glenn M. Lewis - 07/07/92
 */

static char rcs_id[] = "$Id: writedxf.c,v 1.1 1992/07/29 20:46:07 glewis Exp $";

#include <stdio.h>
#include <ctype.h>
#include "ttdddlib.h"

static void process_DESC();
static void process_INFO();
static FILE *out;

/* Here are a few necessary utilities */

static void send_fXYZ(i, f1, f2, f3)			/* Print a common string */
int i;
int f1,f2,f3;
{
	//Face 1:    A:1 B:3 C:2 AB:1 BC:1 CA:1

	fprintf(out, "Face: %d:  A: %d B: %d C: %d\n\r", i, f1, f2, f3);
}

static void send_XYZ(i, f)			/* Print a common string */
int i;
XYZ_st *f;
{
	fprintf(out, "Vertex: %d:  X: %g ",i, f->val[0]);
	fprintf(out, " Y: %g", f->val[1]);
	fprintf(out, " Z: %g\n\r",  f->val[2]);
}

static void send_RGB(rgb)			/* Print a common string */
RGB_st *rgb;
{
	fprintf(out, "%g\t%g\t%g\n\r", ((double)rgb->val[0])/255.0,
		((double)rgb->val[1])/255.0,
		((double)rgb->val[2])/255.0);
}

/********************/
/* The MAIN section */
/********************/

int write_ASC(world, file)
WORLD *world;
FILE *file;
{
	register OBJECT *o;

	if (!(out = file) || !world) return(0);

	/* Write header to DXF file: */
//	fprintf(out, "  0\nSECTION\n  2\nENTITIES\n");

	if (world->info) process_INFO(world->info);

	for (o=world->object; o; o=o->next)
		if (!o->extr) process_DESC(o);

//	fprintf(out, "  0\nENDSEC\n  0\nEOF\n");

	return(1);
}

output_DESC(object)
OBJECT *object;
	{
	process_DESC(object);
	}

static void process_DESC(object)
OBJECT *object;
{
	register int i;
	register OBJECT *obj;
	register DESC *desc = object->desc;
	register int p1, p2, p3;
	int vcount;

	/* Process children first */
	for (obj=object->child; obj; obj=obj->next) {
		if (!obj->extr) process_DESC(obj);
	}

	if (!desc->pcount || !desc->fcount) return;

/*	fprintf(out, "\n# Start of new object\n");	*/
	fprintf(out,"Ambient light color: Red=0.3 Green=0.3 Blue=0.3\n\r");
	fprintf(out,"Named object: \"%s\"\n\r",desc->name);
	printf("Named object: \"%s\"\n\r",desc->name);
//Tri-mesh, Vertices: 298     Faces: 458
	fprintf(out,"Tri-mesh, Vertices: %d  Faces: %d\n\r", desc->fcount*3,desc->fcount);
	printf("Tri-mesh, Vertices: %d  Faces: %d\n\r", desc->fcount*3,desc->fcount);
	fprintf(out,"Vertex list:\r");
	vcount=0;
	for (i=0; i<desc->fcount; i++)
		{
	/* First check to make sure that this triangle is real */
		p1 = desc->edge[desc->face[i*3]];
		p2 = desc->edge[desc->face[i*3]+1];
		p3 = desc->edge[desc->face[i*3]+2];
		p1 = desc->edge[(desc->face[i*3])<<1];
		p2 = desc->edge[((desc->face[i*3])<<1)+1];
//		if (p1 == p2) continue;	/* How did *this* happen? */
		p3 = desc->edge[(desc->face[i*3+2])<<1];
		if (p1 == p3 || p2 == p3)
			p3 = desc->edge[((desc->face[i*3+2])<<1)+1];
		if (p1 == p3 || p2 == p3) continue; /* How did *this* happen? */
		/* Now check the actual points for equality */
/*		if (bcmp(&desc->pnts[p1], &desc->pnts[p2], sizeof(XYZ_st))==0 ||
			bcmp(&desc->pnts[p1], &desc->pnts[p3], sizeof(XYZ_st))==0 ||
			bcmp(&desc->pnts[p2], &desc->pnts[p3], sizeof(XYZ_st))==0)
			continue;*/

//		fprintf(out, "  0\n3DFACE\n  8\n0\n  6\nCONTINUOUS\n 62\n     4\n");
		send_XYZ(vcount, &desc->pnts[p1]);
		vcount++;
		send_XYZ(vcount, &desc->pnts[p2]);
		vcount++;
		send_XYZ(vcount, &desc->pnts[p3]);
		vcount++;
		}
	vcount=0;
	fprintf(out,"Face list:\n\r");
	for (i=0; i<desc->fcount; i++) {
	/* Output the Faces */

//		fprintf(out, "  0\n3DFACE\n  8\n0\n  6\nCONTINUOUS\n 62\n     4\n");
		send_fXYZ(i, vcount, vcount+1, vcount+2);
		vcount+=3;
	}
}

static void process_INFO(info)
INFO *info;
{
#ifdef OLD_NFF_CODE
	fprintf(out, "v\n");
	if (info->obsv) {
		fprintf(out, "from "); send_XYZ(&info->obsv->came);
	}
	if (info->ambi)
		{ fprintf(out, "b "); send_RGB(info->ambi); }
#endif
}
